"""Tests for `utils`."""

import unittest

import jax
import numpy as onp
import parameterized

from agjax import utils


class WrappedValueTest(unittest.TestCase):
    def test_flatten_unflatten(self):
        wrapped = utils.WrappedValue(value=(1, 2, 3, 4))
        leaves, treedef = jax.tree_util.tree_flatten(wrapped)
        self.assertSequenceEqual(leaves, ())
        restored = jax.tree_util.tree_unflatten(treedef, leaves)
        self.assertSequenceEqual(restored.value, (1, 2, 3, 4))

    def test_wrapped_repr(self):
        wrapped = utils.WrappedValue(value="my_string")
        self.assertSequenceEqual(str(wrapped), "_WrappedValue(my_string)")


class ValidateIdxTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        (
            ([6], 3),
            ([-4], 3),
            ([-4, -3, -2], 3),
        )
    )
    def test_invalid_idx_out_of_bounds(self, idx, sequence_length):
        with self.assertRaisesRegex(ValueError, "Found out of bounds values"):
            utils.validate_idx_for_sequence_len(idx, sequence_length)

    @parameterized.parameterized.expand(
        (
            ([1, 1], 3),
            ([1, -2], 3),
        )
    )
    def test_invalid_idx_duplicate_value(self, idx, sequence_length):
        with self.assertRaisesRegex(ValueError, "Found duplicate values"):
            utils.validate_idx_for_sequence_len(idx, sequence_length)


class SplitMergeTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        (
            ([1, 2, 3, 4], [0, 1]),
            ([1, 2, 3, 4], [0, -1]),
        )
    )
    def test_merge_undoes_split(self, sequence, idx):
        a, b = utils.split(sequence, idx)
        merged = utils.merge(a, b, idx)
        for s, m in zip(sequence, merged):
            onp.testing.assert_array_equal(s, m)



"""agjax - a jax wrapper for autograd-differentiable functions."""

__version__ = "v0.3.7"
__author__ = "Martin Schubert <mfschubert@gmail.com>"

__all__ = ["experimental", "wrap_for_jax"]

from agjax import experimental
from agjax.wrapper import wrap_for_jax



"""Defines a jax wrapper for autograd-differentiable functions.

Copyright (c) 2025 invrs.io LLC
"""

import functools
from typing import Any, Callable, Tuple, Union

import autograd  # type: ignore[import-untyped]
import jax
import numpy as onp
from jax import tree_util

from agjax import utils


def wrap_for_jax(
    fn: Callable[[Any], Any],
    nondiff_argnums: Union[int, Tuple[int, ...]] = (),
    nondiff_outputnums: Union[int, Tuple[int, ...]] = (),
) -> Callable[[Any], Any]:
    """Wraps `fn` so that it can be differentiated by jax.

    Arguments should be jax types, and are converted to numpy arrays prior
    to calling the underlying autograd-differentiable `fn`. Optionally,
    nondifferentiable arguments (i.e. those which cannot be differentiated
    with respect to) may be specified; these are passed to `fn` unchanged.

    Similarly, differentiable outputs are converted to jax types; some
    outputs can be identified as non-differentiable, which are returned
    unchanged.

    Args:
        fn: The autograd-differentiable function.
        nondiff_argnums: The arguments that cannot be differentiated with
            respect to. These are passed to `fn` unchanged.
        nondiff_outputnums: The outputs that cannot be differentiated.
            These are returned exactly as returned by `fn`.

    Returns:
        The wrapped function.
    """
    _nondiff_argnums, _ = utils.ensure_tuple(nondiff_argnums)
    _nondiff_outputnums, _ = utils.ensure_tuple(nondiff_outputnums)
    del nondiff_argnums, nondiff_outputnums

    split_args_fn = functools.partial(utils.split, idx=_nondiff_argnums)
    merge_args_fn = functools.partial(utils.merge, idx=_nondiff_argnums)
    split_outputs_fn = functools.partial(utils.split, idx=_nondiff_outputnums)
    merge_outputs_fn = functools.partial(utils.merge, idx=_nondiff_outputnums)

    @functools.partial(jax.custom_vjp, nondiff_argnums=_nondiff_argnums)
    def _fn(*args_jax: Any) -> Any:
        # Arguments that can be differentiated with respect to are jax arrays, and
        # must be converged to numpy. Extract these, convert to numpy, and remerge.
        nondiff_args, diff_args = split_args_fn(args_jax)
        args = merge_args_fn(nondiff_args, utils.to_numpy(diff_args))
        outputs = fn(*args)
        utils.validate_nondiff_outputnums_for_outputs(_nondiff_outputnums, outputs)
        # Convert differentiable outputs to jax arrays.
        outputs, is_tuple_outputs = utils.ensure_tuple(outputs)
        nondiff_outputs, diff_outputs = split_outputs_fn(outputs)
        nondiff_outputs = tuple([utils.WrappedValue(o) for o in nondiff_outputs])
        outputs = merge_outputs_fn(nondiff_outputs, utils.to_jax(diff_outputs))
        return outputs if is_tuple_outputs else outputs[0]

    def _fwd_fn(*args_jax: Any) -> Any:
        # Convert to numpy the args that can be differentiated with respect to.
        nondiff_args, diff_args = split_args_fn(args_jax)
        args = merge_args_fn(nondiff_args, utils.to_numpy(diff_args))

        # Variables updated nonlocally where `fn` is evaluated.
        is_tuple_outputs: bool = None  # type: ignore[assignment]
        nondiff_outputs: Tuple[Any, ...] = None  # type: ignore[assignment]
        diff_outputs_treedef: tree_util.PyTreeDef = None  # type: ignore[assignment]

        def _tuple_fn(*args: Any) -> onp.ndarray:
            nonlocal is_tuple_outputs
            nonlocal nondiff_outputs
            nonlocal diff_outputs_treedef

            outputs = fn(*args)
            utils.validate_nondiff_outputnums_for_outputs(_nondiff_outputnums, outputs)

            outputs, is_tuple_outputs = utils.ensure_tuple(outputs)
            nondiff_outputs, diff_outputs = split_outputs_fn(outputs)
            nondiff_outputs = utils.arraybox_to_numpy(nondiff_outputs)
            nondiff_outputs = tuple(
                [utils.WrappedValue(o) for o in nondiff_outputs or []]
            )
            diff_outputs_leaves, diff_outputs_treedef = tree_util.tree_flatten(
                diff_outputs
            )
            return autograd.builtins.tuple(tuple(diff_outputs_leaves))

        diff_argnums = tuple(i for i in range(len(args)) if i not in _nondiff_argnums)
        tuple_vjp_fn, diff_outputs_leaves = autograd.make_vjp(
            _tuple_fn, argnum=diff_argnums
        )(*args)
        diff_outputs = tree_util.tree_unflatten(
            diff_outputs_treedef, diff_outputs_leaves
        )
        outputs = merge_outputs_fn(nondiff_outputs, utils.to_jax(diff_outputs))
        outputs = outputs if is_tuple_outputs else outputs[0]

        def _vjp_fn(*diff_outputs: Any) -> Any:
            diff_outputs_leaves = tree_util.tree_leaves(diff_outputs)
            grad = tuple_vjp_fn(utils.to_numpy(diff_outputs_leaves))
            # Note that there is no value associated with nondifferentiable
            # arguments in the return of the vjp function.
            return utils.to_jax(grad)

        return outputs, tree_util.Partial(_vjp_fn)

    def _bwd_fn(*bwd_args: Any) -> Any:
        # The `bwd_args` consist of the nondifferentiable arguments, the
        # residual of the forward function (i.e. our `vjp_fn`), and the
        # vector for which the vector-jacobian product is sought.
        vjp_fn = bwd_args[len(_nondiff_argnums)]
        diff_outputs = bwd_args[len(_nondiff_argnums) + 1 :]
        return vjp_fn(*diff_outputs)

    _fn.defvjp(_fwd_fn, _bwd_fn)

    def _fn_with_unwrapped_outputs(*args_jax: Any) -> Any:
        utils.validate_idx_for_sequence_len(_nondiff_argnums, len(args_jax))
        # Wrapped version of our function with custom vjp, which unpacks the
        # wrapped values associated with nondifferentiable outputs.
        outputs = _fn(*args_jax)
        if not isinstance(outputs, tuple):
            return outputs
        return tuple(
            [o.value if isinstance(o, utils.WrappedValue) else o for o in outputs]
        )

    return _fn_with_unwrapped_outputs



"""Defines a jax wrapper for autograd-differentiable functions.

Copyright (c) 2025 invrs.io LLC
"""

import functools
from packaging import version
from typing import Any, Callable, List, Tuple, Union

import autograd  # type: ignore[import-untyped]
import jax
import jax.numpy as jnp
import numpy as onp

from agjax import utils

if version.Version(jax.__version__) > version.Version("0.4.31"):
    callback_sequential = functools.partial(jax.pure_callback, vmap_method="sequential")
else:
    callback_sequential = functools.partial(jax.pure_callback, vectorized=False)

_FORWARD_STAGE = "fwd"
_BACKWARD_STAGE = "bwd"


def wrap_for_jax(
    fn: Callable[[Any], Any],
    result_shape_dtypes: Any,
    nondiff_argnums: Union[int, Tuple[int, ...]] = (),
    nondiff_outputnums: Union[int, Tuple[int, ...]] = (),
) -> Callable[[Any], Any]:
    """Wraps `fn` so that it can be differentiated by jax.

    The wrapped function is suitable for jax transformations such as `grad`, `jit`,
    `vmap`, and `jacrev`, which is achieved using `jax.pure_callback`.

    Arguments to `fn` must be convertible to jax types, as must all outputs. The
    arguments to the wrapped function should be jax types, and the outputs will be
    jax types.

    Arguments which need not be differentiated with respect to may be specified in
    `nondiff_argnums`, while outputs that need not be differentiated may be specified
    in `nondiff_outputnums`.

    Args:
        fn: The autograd-differentiable function.
        result_shape_dtypes: A pytree matching the jax-converted output of `fn`.
            Specifically, the pytree structure, leaf shapes, and datatypes must match.
        nondiff_argnums: The arguments that cannot be differentiated with respect to.
        nondiff_outputnums: The outputs that cannot be differentiated.

    Returns:
        The wrapped function.
    """
    _nondiff_argnums, _ = utils.ensure_tuple(nondiff_argnums)
    _nondiff_outputnums, _ = utils.ensure_tuple(nondiff_outputnums)
    del nondiff_argnums, nondiff_outputnums

    split_args_fn = functools.partial(utils.split, idx=_nondiff_argnums)
    merge_args_fn = functools.partial(utils.merge, idx=_nondiff_argnums)
    split_outputs_fn = functools.partial(utils.split, idx=_nondiff_outputnums)
    merge_outputs_fn = functools.partial(utils.merge, idx=_nondiff_outputnums)

    # Vjp functions created in the "forward stage" of the calculation, and stored in
    # the `vjp_fns` list. When the calculation switches from the backward to the
    # forward stage, the list of vjp functions is cleared.
    stage = _FORWARD_STAGE
    vjp_fns: List[Callable[[Any], Any]] = []

    @jax.custom_vjp  # type: ignore[misc]
    def _fn(*args: Any) -> Any:
        utils.validate_nondiff_argnums_for_args(_nondiff_argnums, args)
        outputs = callback_sequential(
            lambda *args: fn(*utils.to_numpy(args)),
            result_shape_dtypes,
            *args,
        )
        utils.validate_nondiff_outputnums_for_outputs(_nondiff_outputnums, outputs)
        return outputs

    def _fwd_fn(*args: Any) -> Any:
        def make_vjp(*args: Any) -> Any:
            args = utils.to_numpy(args)
            nonlocal stage
            if stage == _BACKWARD_STAGE:
                vjp_fns.clear()
            stage = _FORWARD_STAGE

            # Variables updated nonlocally where `fn` is evaluated.
            is_tuple_outputs: bool = None  # type: ignore[assignment]
            nondiff_outputs: Tuple[Any, ...] = None  # type: ignore[assignment]
            diff_outputs_treedef: Any = None

            def _tuple_fn(*args: Any) -> onp.ndarray:
                nonlocal is_tuple_outputs
                nonlocal nondiff_outputs
                nonlocal diff_outputs_treedef

                utils.validate_nondiff_argnums_for_args(_nondiff_argnums, args)
                outputs = fn(*args)
                utils.validate_nondiff_outputnums_for_outputs(
                    _nondiff_outputnums, outputs
                )

                outputs, is_tuple_outputs = utils.ensure_tuple(outputs)
                nondiff_outputs, diff_outputs = split_outputs_fn(outputs)
                nondiff_outputs = utils.arraybox_to_numpy(nondiff_outputs)
                diff_outputs_leaves, diff_outputs_treedef = utils.tree_flatten(
                    diff_outputs
                )
                return autograd.builtins.tuple(tuple(diff_outputs_leaves))

            args = utils.to_numpy(args)
            diff_argnums = tuple(
                i for i in range(len(args)) if i not in _nondiff_argnums
            )
            tuple_vjp_fn, diff_outputs_leaves = autograd.make_vjp(
                _tuple_fn, argnum=diff_argnums
            )(*args)
            diff_outputs = utils.tree_unflatten(
                diff_outputs_treedef, diff_outputs_leaves
            )
            outputs = merge_outputs_fn(nondiff_outputs, diff_outputs)
            outputs = outputs if is_tuple_outputs else outputs[0]

            def _vjp_fn(*diff_outputs: Any) -> Any:
                diff_outputs_leaves = utils.tree_leaves(diff_outputs)
                grad = tuple_vjp_fn(utils.to_numpy(diff_outputs_leaves))
                return grad

            key = len(vjp_fns)
            vjp_fns.append(_vjp_fn)
            return outputs, key

        outputs, key = callback_sequential(
            make_vjp,
            (result_shape_dtypes, jnp.asarray(0)),
            *args,
        )
        return outputs, (args, key)

    def _bwd_fn(*bwd_args: Any) -> Any:
        def _pure_fn(key: jnp.ndarray, tangents: Tuple[Any, ...]) -> Any:
            tangents = utils.to_numpy(tangents)
            nonlocal stage
            stage = _BACKWARD_STAGE
            vjp_fn = vjp_fns[int(onp.asarray(key))]
            return vjp_fn(*tangents)

        (args, key), *tangents = bwd_args
        _, diff_args = split_args_fn(args)
        result_shape_dtypes = utils.to_jax(diff_args)
        grads = callback_sequential(
            _pure_fn,
            result_shape_dtypes,
            key,
            tangents,
        )
        return merge_args_fn([None] * len(_nondiff_argnums), grads)

    _fn.defvjp(_fwd_fn, _bwd_fn)

    return _fn  # type: ignore[no-any-return]



# type: ignore
"""Tests for `wrapper`."""

import unittest

import autograd
import autograd.numpy as npa
import jax
import jax.numpy as jnp
import numpy as onp
import parameterized

from agjax import wrapper

TEST_FNS_AND_ARGS = (
    (  # Basic scalar-valued function, real outputs.
        lambda x: x**2,
        (3.0,),
    ),
    (  # Scalar-valued function with two input arguments, real outputs.
        lambda x, y: x**2 + y,
        (3.0, 4.0),
    ),
    (  # Two arguments, scalar output wrapped in a tuple, real outputs.
        lambda x, y: (x**2 + y,),
        (3.0, 4.0),
    ),
    (  # Two arguments, two outputs, real outputs.
        lambda x, y: (x**2 + y, x - y),
        (3.0, 4.0),
    ),
    (  # Two arguments, two outputs, real arguments, complex outputs.
        lambda x, y: (x**2 + 1j * y**2, 1j * x**2 - y**2),
        (3.0, 4.0),
    ),
    (  # Two arguments, two outputs, complex arguments, real outputs.
        lambda x, y: (npa.abs(x) ** 2 + npa.abs(y), npa.abs(x + y)),
        (3.0 + 1.0j, 4.0 + 0.5j),
    ),
    (  # Two arguments, two outputs, complex outputs.
        lambda x, y: (x**2 + y, x - y),
        (3.0 + 1.0j, 4.0 + 0.5j),
    ),
    (  # Two arguments, two outputs, complex outputs.
        lambda x, y, z: (x**2 + y + z, x - y),
        (3.0 + 1.0j, 4.0 + 0.5j, -11.0),
    ),
    (  # Returns a pytree, complex outputs.
        lambda x, y: {"a": x**2 + y, "b": (x - y, y - x)},
        (3.0 + 1.0j, 4.0 + 0.5j),
    ),
    (  # Arguments and outputs include pytree, complex outputs.
        lambda x, y: {
            "a": (x["a0"] + x["a1"]) ** 2 + y,
            "b": (x["a0"] - y, y - x["a1"]),
        },
        ({"a0": 3.0 + 1.0j, "a1": 22.0j}, 4.0 + 0.5j),
    ),
)


class WrapperTest(unittest.TestCase):
    @parameterized.parameterized.expand(([2], [-3]))
    def test_out_of_bounds_nondiff_argnums(self, nondiff_argnums):
        def fn(x, y):
            return (npa.sum(x + y), y)

        wrapped = wrapper.wrap_for_jax(fn, nondiff_argnums=nondiff_argnums)
        with self.assertRaisesRegex(ValueError, "Found out of bounds"):
            wrapped(1.0, 2.0)

    @parameterized.parameterized.expand(([(1, 1)], [(1, -1)]))
    def test_duplicate_nondiff_argnums(self, nondiff_argnums):
        def fn(x, y):
            return (npa.sum(x + y), y)

        wrapped = wrapper.wrap_for_jax(fn, nondiff_argnums=nondiff_argnums)
        with self.assertRaisesRegex(ValueError, "Found duplicate"):
            wrapped(1.0, 2.0)

    @parameterized.parameterized.expand(([2], [-3]))
    def test_out_of_bounds_nondiff_outputnums(self, nondiff_outputnums):
        def fn(x, y):
            return (npa.sum(x + y), y)

        wrapped = wrapper.wrap_for_jax(fn, nondiff_outputnums=nondiff_outputnums)
        with self.assertRaisesRegex(ValueError, "Found out of bounds"):
            wrapped(1.0, 2.0)

    @parameterized.parameterized.expand(([(1, 1)], [(1, -1)]))
    def test_duplicate_nondiff_outputnums(self, nondiff_outputnums):
        def fn(x, y):
            return (npa.sum(x + y), y)

        wrapped = wrapper.wrap_for_jax(fn, nondiff_outputnums=nondiff_outputnums)
        with self.assertRaisesRegex(ValueError, "Found duplicate"):
            wrapped(1.0, 2.0)

    def test_function_has_no_differentiable_outputs(self):
        def fn(x, y):
            return (x, y)

        wrapped = wrapper.wrap_for_jax(fn, nondiff_outputnums=(0, 1))
        with self.assertRaisesRegex(ValueError, "At least one differentiable output"):
            wrapped(1.0, 2.0)
        with self.assertRaisesRegex(ValueError, "At least one differentiable output"):
            jax.grad(wrapped)(1.0, 2.0)
        with self.assertRaisesRegex(ValueError, "At least one differentiable output"):
            jax.value_and_grad(wrapped)(1.0, 2.0)

    @parameterized.parameterized.expand(TEST_FNS_AND_ARGS)
    def test_wrapped_matches_autograd(self, autograd_fn, args):
        # Tests case where all arguments can be differentiated
        # with respect to, and all outputs are differentiable.
        expected_outputs = autograd_fn(*args)

        wrapped = wrapper.wrap_for_jax(
            autograd_fn,
            nondiff_argnums=(),
            nondiff_outputnums=(),
        )
        for v, ev in zip(
            jax.tree_util.tree_leaves(wrapped(*args)),
            jax.tree_util.tree_leaves(expected_outputs),
        ):
            onp.testing.assert_allclose(v, ev)

        def autograd_scalar_fn(*args):
            outputs = autograd_fn(*args)
            outputs = jax.tree_util.tree_leaves(outputs)
            return npa.sum([npa.sum(npa.abs(o) ** 2) for o in outputs])

        def jax_scalar_fn(*args):
            outputs = wrapped(*args)
            outputs = jax.tree_util.tree_leaves(outputs)
            return jnp.sum(jnp.asarray([jnp.sum(jnp.abs(o) ** 2) for o in outputs]))

        expected_grad = autograd.grad(autograd_scalar_fn)(*args)
        grad = jax.grad(jax_scalar_fn)(*args)

        for g, eg in zip(
            jax.tree_util.tree_leaves(grad), jax.tree_util.tree_leaves(expected_grad)
        ):
            onp.testing.assert_allclose(g, eg)

    def test_nondiff_argnums_and_outputnums(self):
        def autograd_fn(x, y, int_arg, str_arg):
            return (
                npa.sum(x**2 + y * int_arg),
                x - y + 2,
                str_arg * int_arg,
            )

        wrapped = wrapper.wrap_for_jax(
            autograd_fn, nondiff_argnums=(2, 3), nondiff_outputnums=2
        )

        args = (0.3 + 2.2j, -11.0 + 0.0j, 3, "test")
        expected_outputs = ((0.3 + 2.2j) ** 2 - 11.0 * 3, 13.3 + 2.2j, "testtesttest")

        # Check that directly calling the wrapped function gives the expected
        # output. This validates `_fn`.
        wrapped_outputs = wrapped(*args)
        onp.testing.assert_allclose(wrapped_outputs[0], expected_outputs[0])
        onp.testing.assert_allclose(wrapped_outputs[1], expected_outputs[1])
        self.assertEqual(wrapped_outputs[2], expected_outputs[2])

        def jax_scalar_fn(*args):
            a, b, c = wrapped(*args)
            return a + b, c

        expected_value = (0.3 + 2.2j) ** 2 - 11.0 * 3 + 13.3 + 2.2j
        expected_aux = "testtesttest"
        expected_grad = (2 * args[0] + 1, 2)

        # Check that directly calling the wrapped function gives the expected
        # output. This validates `_fn`.
        value, aux = jax_scalar_fn(*args)
        onp.testing.assert_allclose(value, expected_value)
        self.assertEqual(aux, expected_aux)

        # Check that `value_and_grad` returns the correct value and gradient.
        # The value is computed by `fwd_fn`, and so this tests a different
        # codepath than when the value is computed directly.
        (value, aux), grad = jax.value_and_grad(
            jax_scalar_fn,
            has_aux=True,
            holomorphic=True,
            argnums=(0, 1),
        )(*args)
        onp.testing.assert_allclose(value, expected_value)
        self.assertEqual(aux, expected_aux)
        for e, g in zip(expected_grad, grad):
            onp.testing.assert_allclose(e, g)



"""Defines a utility functions for jax-autograd wrappers.

Copyright (c) 2025 invrs.io LLC
"""

from typing import Any, Callable, Sequence, Tuple

import autograd.numpy as npa  # type: ignore[import-untyped]
import jax.numpy as jnp
import numpy as onp
from jax import tree_util

PyTree = Any


class WrappedValue:
    """Wraps a value treated as an auxilliary quantity of a pytree node."""

    def __init__(self, value: Any) -> None:
        self.value = value

    def __repr__(self) -> str:
        return f"_WrappedValue({self.value})"


tree_util.register_pytree_node(
    WrappedValue,
    flatten_func=lambda w: ((), (w.value,)),
    unflatten_func=lambda v, _: WrappedValue(*v),
)


def validate_nondiff_outputnums_for_outputs(
    nondiff_outputnums: Sequence[int],
    maybe_tuple_outputs: Any,
) -> None:
    """Validates that `nondiff_outputnums` is compatible with a `outputs`."""
    outputs_length = (
        len(maybe_tuple_outputs) if isinstance(maybe_tuple_outputs, tuple) else 1
    )
    validate_idx_for_sequence_len(nondiff_outputnums, outputs_length)
    if outputs_length <= len(nondiff_outputnums):
        raise ValueError(
            f"At least one differentiable output is required, but got "
            f"`nondiff_outputnums` of {nondiff_outputnums} when `fn` "
            f"has {outputs_length} output(s)."
        )


def validate_nondiff_argnums_for_args(
    nondiff_argnums: Sequence[int],
    args: Tuple[Any, ...],
) -> None:
    """Validates that `nondiff_argnums` is compatible with a `args`."""
    validate_idx_for_sequence_len(nondiff_argnums, len(args))
    if len(args) <= len(nondiff_argnums):
        raise ValueError(
            f"At least argument must be differentiated with respect to, but got "
            f"`nondiff_argnums` of {nondiff_argnums} when `fn` has {len(args)} "
            f"arguments(s)."
        )


def validate_idx_for_sequence_len(idx: Sequence[int], sequence_length: int) -> None:
    """Validates that `idx` is compatible with a sequence length."""
    if not all(i in range(-sequence_length, sequence_length) for i in idx):
        raise ValueError(
            f"Found out of bounds values in `idx`, got {idx} when "
            f"`sequence_length` is {sequence_length}."
        )
    positive_idx = [i % sequence_length for i in idx]
    if len(positive_idx) != len(set(positive_idx)):
        raise ValueError(
            f"Found duplicate values in `idx`, got {idx} when "
            f"`sequence_length` is {sequence_length}."
        )


def split(
    a: Tuple[Any, ...],
    idx: Tuple[int, ...],
) -> Tuple[Tuple[Any, ...], Tuple[Any, ...]]:
    """Splits the sequence `a` into two sequences."""
    validate_idx_for_sequence_len(idx, len(a))
    return (
        tuple([a[i] for i in idx]),
        tuple([a[i] for i in range(len(a)) if i not in idx]),
    )


def merge(
    a: Sequence[Any],
    b: Sequence[Any],
    idx: Sequence[int],
) -> Tuple[Any, ...]:
    """Merges the sequences `a` and `b`, undoing a `_split` operation."""
    validate_idx_for_sequence_len(idx, len(a) + len(b))
    positive_idx = [i % (len(a) + len(b)) for i in idx]
    iter_a = iter(a)
    iter_b = iter(b)
    return tuple(
        [
            next(iter_a) if i in positive_idx else next(iter_b)
            for i in range(len(a) + len(b))
        ]
    )


def to_jax(tree: PyTree) -> PyTree:
    """Converts leaves of a pytree to jax arrays."""
    return tree_map(jnp.asarray, tree)


def to_numpy(tree: PyTree) -> PyTree:
    """Converts leaves of a pytree to numpy arrays."""
    return tree_map(onp.asarray, tree)


def arraybox_to_numpy(tree: PyTree) -> PyTree:
    """Converts `ArrayBox` leaves of a pytree to numpy arrays."""
    return tree_map(
        lambda x: x._value if isinstance(x, npa.numpy_boxes.ArrayBox) else x,
        tree,
    )


def ensure_tuple(xs: Any) -> Tuple[Any, bool]:
    """Returns `(xs, True)` if `xs` is a tuple, and `((xs,), False)` otherwise."""
    is_tuple = isinstance(xs, tuple)
    return (xs if is_tuple else (xs,)), is_tuple


# -----------------------------------------------------------------------------
# Non-jax versions of some `jax.tree_util` functions.
# -----------------------------------------------------------------------------


def tree_flatten(tree: PyTree) -> Tuple[Any, PyTree]:
    """Pure-python version of `jax.tree_util.tree_flatten`."""
    leaves = []

    def _flatten(tree):
        if not _is_node(tree):
            assert tree is not None  # Cannot handle `None` leaves.
            leaves.append(tree)
            return None
        elif isinstance(tree, dict):
            return {k: _flatten(v) for k, v in tree.items()}
        else:
            return type(tree)([_flatten(v) for v in tree])

    treedef = _flatten(tree)
    return leaves, treedef


def tree_unflatten(treedef, leaves):
    """Pure-python version of `jax.tree_util.tree_unflatten`."""
    leaves = iter(leaves)

    def _unflatten(tree):
        if not _is_node(tree):
            return next(leaves)
        elif isinstance(tree, dict):
            return {k: _unflatten(v) for k, v in tree.items()}
        else:
            return type(tree)([_unflatten(v) for v in tree])

    return _unflatten(treedef)


def tree_map(fn: Callable[[Any], Any], tree: PyTree) -> PyTree:
    """Pure-python version of `jax.tree_util.tree_map`."""
    leaves, treedef = tree_flatten(tree)
    return tree_unflatten(treedef, [fn(leaf) for leaf in leaves])


def tree_leaves(tree: PyTree) -> Any:
    """Pure-python version of `jax.tree_util.tree_leaves`."""
    leaves, _ = tree_flatten(tree)
    return leaves


def _is_node(maybe_node: Any) -> bool:
    """Return `True` if `maybe_node` is actually a node."""
    return isinstance(maybe_node, (tuple, list, dict))
